#  SPDX-FileCopyrightText: © 2025 Josef Hahn
#  SPDX-License-Identifier: AGPL-3.0-only
import abc
import asyncio
import typing as t

import klovve.app.runnable
import klovve.data
import klovve.driver
import klovve.event
import klovve.builtin.views.window
import klovve.builtin.views.interact
import klovve.object
import klovve.timer
import klovve.variable


class BaseApplication(abc.ABC):

    class ActionTriggeredEvent(klovve.event.Event):

        def __init__(self, triggering_view: klovve.ui.View, action_name: str):
            super().__init__()
            self.__triggering_view = triggering_view
            self.__action_name = action_name

        @property
        def action_name(self) -> str:
            return self.__action_name

        @property
        def triggering_view(self) -> klovve.ui.View:
            return self.__triggering_view

    @property
    @abc.abstractmethod
    def windows(self) -> klovve.data.list.List["klovve.builtin.views.window.Window"]:
        pass

    @abc.abstractmethod
    async def dialog(self, dialog: type["klovve.ui.dialog.BaseDialog"]|klovve.builtin.views.interact._AbstractInteract,
                     dialog_args=None, dialog_kwargs=None, *, view_anchor) -> t.Any:
        pass

    @abc.abstractmethod
    def _handle_event(self, event):
        pass

    @property
    @abc.abstractmethod
    def _is_running(self) -> bool:
        pass

    @abc.abstractmethod
    async def _wait_until_finished(self) -> None:
        pass

    @property
    @abc.abstractmethod
    def driver_compatibility(self) -> "klovve.driver.Driver.CompatibilitySpecification":
        pass

    @abc.abstractmethod
    def _started(self, driver: "klovve.driver.Driver") -> None:
        pass

    def enqueue_task(self, coro):
        klovve.driver.Driver.get().loop.enqueue(coro)

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class Application(klovve.object.Object, klovve.event._EventHandlingObject, klovve.timer._TimingObject,
                  klovve.object.WithPublicBind, BaseApplication):

    def __init__(self, **kwargs):
        super().__init__()
        self._set_data_by_kwargs(kwargs)
        self.__windows = klovve.data.list.List["klovve.builtin.views.window.Window"]()
        self.__windows.add_observer(
            Application._RemoveWindowsWhenClosedListObserver, (self,),
            initialize_with=klovve.variable.no_dependency_tracking(), owner=self)
        self.__finished_future = None

    def _started(self, driver: "klovve.driver.Driver") -> None:
        self._initialize_timing()
        self.__finished_future = asyncio.Future()
        self.__windows.add_observer(
            Application._FinishApplicationAfterLastWindowClosedListObserver, (self, self.__finished_future),
            initialize_with=klovve.variable.no_dependency_tracking(), owner=self)

    @property
    def driver_compatibility(self):
        return Application._CompatibilitySpecification(level=klovve.driver.Driver.LEVEL_GRAPHICAL)

    @property
    def windows(self):
        return self.__windows

    @property
    def _is_running(self) -> bool:
        return (self.__finished_future is not None) and (not self.__finished_future.done())

    async def dialog(self, dialog, dialog_args=None, dialog_kwargs=None, *, view_anchor):
        if isinstance(dialog, klovve.ui.View):
            dialog = klovve.ui.dialog._InteractDialog[dialog]

        return await klovve.driver.Driver.get().dialog_host.show_dialog(self, view_anchor,
                                                                        dialog, dialog_args, dialog_kwargs)

    async def _wait_until_finished(self):
        await self.__finished_future

    class _CompatibilitySpecification(klovve.driver.Driver.CompatibilitySpecification):

        def __init__(self, *, level: float):
            self.__level = level

        def is_driver_compatible(self, driver):
            return driver.level() >= self.__level

    class _FinishApplicationAfterLastWindowClosedListObserver(klovve.data.list.List.Observer):

        def __init__(self, app, finished_future):
            super().__init__()
            self.__app = app
            self.__finished_future = finished_future

        def item_removed(self, index, item):
            if len(self.__app.windows) == 0:
                self.__finished_future.set_result(True)

        def item_added(self, index, item):
            pass

    class _RemoveWindowsWhenClosedListObserver(klovve.data.list.List.Observer):

        def __init__(self, app):
            super().__init__()
            self.__app = app

        def item_removed(self, index, item):
            pass

        def item_added(self, index, item):
            klovve.effect.activate_effect(Application._RemoveWindowWhenClosedEffect, (self.__app, item), owner=item)

    class _RemoveWindowWhenClosedEffect(klovve.effect.Effect):

        def __init__(self, app, window):
            super().__init__()
            self.__app = app
            self.__window = window

        def run(self):
            if self.__window._is_closed and self.__window in self.__app.windows:
                self.__app.windows.remove(self.__window)


def create(application: t.Union[BaseApplication, "klovve.builtin.views.window.Window"]) -> "klovve.app.runnable.Runnable":
    if isinstance(application, klovve.builtin.views.window.Window):
        window = application
        application = Application()
        application.windows.append(window)

    return _create(application, klovve.driver.Driver.get(application.driver_compatibility))


def _create(application: BaseApplication, driver: "klovve.driver.Driver") -> "klovve.app.runnable.Runnable":
    return driver.new_runnable_for_application(application)
